from sigma import sigma_opt, sigma_new 
import numpy as np


def vander(x,j,tau=1):
    """helper function
    """
    exponents =[p for p in range(len(x)*tau)] 
    x=np.array(x)
    return x[:, None] ** exponents


def sparse_ifft(xd, threshold=0.1, accuracy=2, taumax=3, sigma_function=sigma_opt):
    """reconstruction of x from the Fourier entries FN(x)=xd
    Args:
        xd (1D numpy array):   Fourier Transform of x, vector of length N=2^J    
        threshold (float):     if abs(values) < threshold: value will be treated as zero  
        taumax (integer):      factor to increase number of rows in the vandermonde-matrix
        accuracy (integer):     the higher the more precise the algorithm works

    Output: 
        I (list):   the support indices of x 
        M (integer):    sparsity of x
        Xj (1D numpy array):   the short vector containing only the significant components of x
            
    """

    if abs(xd[0])<=threshold:
        return np.zeros(len(xd), dtype='complex')

    else:
        N=len(xd)
        J=int(np.log2(N))
        I=[0]   
        Xj=[xd[0]] #X^0   
        M=1
        Mold=0  
        L=0.5      #L=2^j
        L1=N       #L1=2**(J-j-1)
        tau=1

        for j in range(J):
            L=int(2*L) 
            L1=int(L1/2)

            if  (M**2>=L or M==0):
                # if Xj=Xshort->expand Xj
                if len(Xj)<L: 
                    Yj=np.zeros(L,dtype='complex')
                    for i in range(M):
                        Yj[I[i]]=Xj[i]
                    Xj=Yj

                # x^(j+1) 
                zd=np.array([xd[L1*(2*i+1)] for i in range(L)]) 
                xf=np.fft.ifft(np.transpose(zd))
                x0=0.5*(Xj+np.array([np.exp(np.pi*1j*k/L)*xf[k] for k in range(L)]))
                x1=Xj-x0
                Xj=[x for x in x0]+[x for x in x1]

                # I^(j+1) 
                Mold=M
                I=[i for i in range(len(x0)) if abs(x0[i])>=threshold]+[i+L for i in range(len(x1)) if abs(x1[i])>=threshold]
                M=len(I)
                last=j

            else:
                if j==(last+1):
                    Perm=np.eye(M) 

                Xj=[round(x0[i].real, 2) + round(x0[i].imag, accuracy) * 1j for i in range(len(x0)) if abs(x0[i])>threshold]+[round(x1[i].real, accuracy) +                                                                                                        round(x1[i].imag, accuracy) * 1j  for i in range(len(x1)) if abs(x1[i])>threshold] 

                #sigma and dsigma 
                if M==1:
                    sigma, dsigma=1,L
                elif M!=Mold or last+1==j:
                    sigma, dsigma=sigma_function(L,M,I)
                else :   
                    sigma=2*sigma 
                    dsigma*=2
                
                # calculate tau and V 
                if (M!=Mold or j==last+1): 
                    if dsigma>0 and M>0:
                        optdist= L/M 
                        tau=min(int(optdist/dsigma),taumax) 
                    xv=[np.exp(-2*np.pi*1j*sigma*i/(L)) for i in I]               
                    V=(np.transpose(vander(xv,j,tau)))
                    if tau>1:
                        Q, R = np.linalg.qr(V) 
                    
                #Permutation 
                if M==Mold and (Perm!=np.eye(M)).any():    
                    if tau==1:
                        V=V@Perm                              
                    else:
                        R=R@Perm                      
                               
                
                # zd^(j+1) 
                Kp=[L1*(2*((sigma*p)%L)+1)for p in range(tau*M)]
                zd=[xd[i] for i in Kp]  

                # x0^(j+1) 
                diagwn=np.diag([(np.exp(np.pi*1j*i/L)) for i in I])
                if tau>1:
                    z=np.transpose(np.conj(Q))@zd
                    x0=0.5*(diagwn@np.linalg.inv(R)@z+Xj)
                else: 
                    x0=0.5*(diagwn@np.linalg.inv(V)@zd+Xj)

                # x^(j+1) & I^(j+1) 
                x1=Xj-x0   
                Mold=M
                Iold=I
                I=[I[i] for i in range(len(x0)) if abs(x0[i])>=threshold ]+[I[i]+L for i in range(len(x1)) if abs(x1[i])>=threshold ]
                Xj=[round(x0[i].real, accuracy) + round(x0[i].imag, accuracy) * 1j for i in range(len(x0)) if abs(x0[i])>threshold]+[round(x1[i].real, accuracy) + round(x1[i].imag, accuracy) * 1j  for i in range(len(x1)) if abs(x1[i])>threshold]                                       
                M=len(I)
            

                # PERM  
                if M==Mold:
                    Perm=np.eye(M)
                    for i in range(M):
                        for k in range(M):
                            if k!=i and (Iold[i]==I[k] or Iold[i]+L==I[k]):
                                    Perm[i][i]=0
                                    Perm[i][k]=1
           

        Xj=[round(x0[i].real, accuracy) + round(x0[i].imag, accuracy) * 1j for i in range(len(x0)) if abs(x0[i])>threshold]+[round(x1[i].real, accuracy) + round(x1[i].imag, accuracy) * 1j  for i in range(len(x1)) if abs(x1[i])>threshold]                                       

        return I, M, Xj       

